unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, ExtCtrls, Grids, Params, StdCtrls, Instructions, ChurchCensus, Grid;

type
  TMainScreen = class(TForm)
    MainMenu1: TMainMenu;
    SetParameters: TMenuItem;
    RunOne: TMenuItem;
    KeysPanel: TPanel;
    Exit: TMenuItem;
    DrawGrid1: TDrawGrid;
    Instructions1: TMenuItem;
    N1: TMenuItem;
    N4: TMenuItem;
    ResetSimulation: TMenuItem;
    N3: TMenuItem;
    N2: TMenuItem;
    RunFive: TMenuItem;
    N5: TMenuItem;
    RunTen: TMenuItem;
    N6: TMenuItem;
    SplashPanel: TPanel;
    DescriptionLbl: TLabel;
    TitleLabel: TLabel;
    AuthorLbl: TLabel;
    EmailLbl: TLabel;
    Run50: TMenuItem;
    N7: TMenuItem;
    FillLbl1: TLabel;
    FillLbl2: TLabel;
    CopyrightLbl: TLabel;
    FreewareLbl: TLabel;
    FillLbl3: TLabel;
    FillLbl4: TLabel;
    FillLbl5: TLabel;
    Instructions2: TMenuItem;
    ProgramOptions1: TMenuItem;
    Background1: TMenuItem;
    procedure ExitClick(Sender: TObject);
    procedure SetParametersClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ResetSimulationClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure RunOneClick(Sender: TObject);
    procedure RunFiveClick(Sender: TObject);
    procedure RunTenClick(Sender: TObject);
    procedure Run50Click(Sender: TObject);
    procedure Instructions2Click(Sender: TObject);
    procedure ProgramOptions1Click(Sender: TObject);
    procedure Background1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

type
  TRowArray = array[0..99] of char;
  TGridArray = array[0..99] of TRowArray;
  TConvTendRowArray = array[0..99] of integer;       {added for version 1.5}
  TConvTendArray = array[0..99] of TConvTendRowArray;   {added for version 1.5}
  TOrder = array[1..10000] of integer;
  TColorCode = array[0..11] of char;
  TLgOrSmArray = array[0..11] of char;
  TCloseNeighborsArray = array[0..7] of char;
  TFarNeighborsArray = array[0..15] of char;

const
  TotalNum = 10000;
  OrgScreenHeight = 864;
  OrgScreenWidth = 1152;
  clLtBlue = $00FFD6BA;          {light blue}
  clLtYell = $00E2FFFF;         {light yellow}
  clLtBrown = $00ECE9D9;         {replaces buttonface}

var
  MainScreen: TMainScreen;
  ScreenHeight, ScreenWidth : integer;
  PctSecular : integer;  {% of pop that has no religion}
  LgRelAttraction, SmRelAttraction,
     PlausibilityRate, SecularizationRate,
     InvestmentRate, LgInvRate, SmInvRate, NOInvRate : integer;
  NumLgRel, NumSmRel : integer;  {# of each chosen}
  SizeRatio : integer; {relative size of largest to smallest church at start}

  NumReligious : integer;
  UnitSize : integer;  {size of smallest church at start}
  ShiftToLarge, ShiftToSmall, ShiftToMed : integer;
  NoSmall, NoMed, NoLarge : boolean;  {flags for missing churches}
  TreatAsBarrier, IndivVary, StPaulFactor : boolean;

  ChurchSize, StartChurchSize : array[0..11] of integer;
  LgOrSmNow, StartLgOrSmNow : TLgOrSmArray;    {for keeping track of church size}
  ChurchCounter : integer;
  ColorCode : TColorCode;

  GridArray, StartArray, HoldArray, BlankArray : TGridArray;
  ConvTendArray : TConvTendArray;    {added for version 1.5}

  Order : TOrder;

  CloseNeighbors : TCloseNeighborsArray;
  FarNeighbors : TFarNeighborsArray;

  StopHere : boolean;



{----------------------------------IMPLEMENTATION---------------------------}

implementation

uses Settings, ExitQuestion, ErrorBox, Options, Background;

{$R *.DFM}

{------------------------------ INTERNAL PROCEDURES -------------------------}

procedure SeedRandom(var RandomOrder : TOrder);
var        {sets up random order for filling grid}
  I,J,K : integer;
  ReDo : boolean;
begin
  Randomize;
  I := 1;
  while I <= TotalNum do
  begin
    ReDo := false;
    J := Random(TotalNum) + 1;
    for K := 1 to I do if RandomOrder[K] = J then ReDo := true;
    if ReDo = false then
    begin
      RandomOrder[I] := J;
      I := I + 1;
    end;
  end;
end;

procedure ExamineClose(Row, Col : integer;
                            var CloseNeighbors : TCloseNeighborsArray);
var
  CheckUp, CheckDown, CheckLeft, CheckRight : integer;
  I : integer;
begin
     {identify surrounding cells}
  if (Row = 0) then CheckUp := 99 else CheckUp := Row - 1;
  if (Col = 0) then CheckLeft := 99 else CheckLeft := Col - 1;
  if (Row = 99) then CheckDown := 0 else CheckDown := Row + 1;
  if (Col = 99) then CheckRight := 0 else CheckRight := Col + 1;

     {blank array to pass}
  for I := 0 to 7 do
  begin
    CloseNeighbors[I] := 'z';
  end;

     {fill array with surrounding colors}
  CloseNeighbors[0] := GridArray[CheckUp, CheckLeft];
  CloseNeighbors[1] := GridArray[CheckUp, Col];
  CloseNeighbors[2] := GridArray[CheckUp, CheckRight];
  CloseNeighbors[3] := GridArray[Row, CheckLeft];
  CloseNeighbors[4] := GridArray[Row, CheckRight];
  CloseNeighbors[5] := GridArray[CheckDown, CheckLeft];
  CloseNeighbors[6] := GridArray[CheckDown, Col];
  CloseNeighbors[7] := GridArray[CheckDown, CheckRight];
end;

procedure ExamineFar(Row, Col : integer;
                            var FarNeighbors : TFarNeighborsArray);
var
  CheckUp1, CheckUp2, CheckDown1, CheckDown2, CheckLeft1, CheckLeft2,
            CheckRight1, CheckRight2 : integer;
  I : integer;
begin
     {identiry surrounding cells}
  if (Row = 0) then
  begin
    CheckUp2 := 98; CheckUp1 := 99; end
    else if (Row = 1) then
    begin
      CheckUp2 := 99; CheckUp1 := 0; end
    else begin CheckUp2 := Row - 2; CheckUp1 := Row - 1; end;
  if (Col = 0) then
  begin
    CheckLeft2 := 98; CheckLeft1 := 99; end
    else if (Col = 1) then
    begin
      CheckLeft2 := 99; CheckLeft1 := 0; end
      else begin CheckLeft2 := Col - 2; CheckLeft1 := Col - 1; end;
  if (Row = 99) then
  begin
    CheckDown2 := 1; CheckDown1 := 0; end
    else if (Row = 98) then
    begin
      CheckDown2 := 0; CheckDown1 := 99; end
      else begin CheckDown2 := Row + 2; CheckDown1 := Row+1; end;
  if (Col = 99) then
  begin
    CheckRight2 := 1; CheckRight1 := 0; end
    else if (Col = 98) then
    begin
      CheckRight2 := 0; CheckRight1 := 99; end
      else begin CheckRight2 := Col + 2; CheckRight1 := Col + 1; end;

     {blank array to pass}
  for I := 0 to 23 do
  begin
    FarNeighbors[I] := 'z';
  end;

     {fill array with surrounding colors}
  FarNeighbors[0] := GridArray[CheckUp2, CheckLeft2];
  FarNeighbors[1] := GridArray[CheckUp2, CheckLeft1];
  FarNeighbors[2] := GridArray[CheckUp2, Col];
  FarNeighbors[3] := GridArray[CheckUp2, CheckRight1];
  FarNeighbors[4] := GridArray[CheckUp2, CheckRight2];
  FarNeighbors[5] := GridArray[CheckUp1, CheckLeft2];
  FarNeighbors[6] := GridArray[CheckUp1, CheckRight2];
  FarNeighbors[7] := GridArray[Row, CheckLeft2];
  FarNeighbors[8] := GridArray[Row, CheckRight2];
  FarNeighbors[9] := GridArray[CheckDown1, CheckLeft2];
  FarNeighbors[10] := GridArray[CheckDown1, CheckRight2];
  FarNeighbors[11] := GridArray[CheckDown2, CheckLeft2];
  FarNeighbors[12] := GridArray[CheckDown2, CheckLeft1];
  FarNeighbors[13] := GridArray[CheckDown2, Col];
  FarNeighbors[14] := GridArray[CheckDown2, CheckRight1];
  FarNeighbors[15] := GridArray[CheckDown2, CheckRight2];
end;

{----------------------------- REASSIGNMENT PROCEDURES ---------------------}

procedure Reassign;
type
  TNeighborCount = array[0..11] of single;
var
  I, J, K, L, M, N, P, Q, R : integer;
  StPaulOrder : TOrder;
  CloseNeighborCount : TNeighborCount;
  FarNeighborCount : TNeighborCount;
  PlausibilityCount : integer;
  MultFactor : TNeighborCount;
  MostAttraction : single;
  MostAttractCode : integer;
  RandomChoice : integer;
  ChurchSizeNow : char;
  OldColor, NewColor : char;
  InvMultiplier, InvDivider : integer;
  MultipFactor : real;

begin
                           {FOCUS ON EACH CELL SEQUENTIALLY}
  for I := 0 to 99 do
  begin
    for J := 0 to 99 do
    begin
      for K := 0 to 11 do CloseNeighborCount[K] := 0;
      for K := 0 to 11 do FarNeighborCount[K] := 0;

                    {1. GET COUNTS OF NEIGHBORS' RELIGIONS}

      ExamineClose(I,J, CloseNeighbors);
{      ExamineFar(I,J, FarNeighbors);    }
      for K := 0 to 11 do       {determines current size of religion}
      begin
        if GridArray[I,J] = ColorCode[K] then
        begin
          ChurchSizeNow := LgOrSmNow[K];
        end;
        for L := 0 to 7 do
        begin
          if CloseNeighbors[L] = ColorCode[K]then
          begin
            CloseNeighborCount[K] := CloseNeighborCount[K] + 1;
          end;
        end;
{        for M := 0 to 15 do
        begin
          if FarNeighbors[M] = ColorCode[K] then
          begin
             if GridArray[I,J] <> ColorCode[K]
             then FarNeighborCount[K] := FarNeighborCount[K] + 1;
          end;
        end;     }
      end;

                         {2. FIGURE ATTRACTION FACTORS}
                       {LARGE, SMALL, & SECULARIZATION}
      for K := 0 to 11 do
      begin
        MultFactor[K] := 0;
        if CloseNeighborCount[K] <> 0 then
        begin
          if K = 0 then
          begin
            MultFactor[K] := CloseNeighborCount[K]*(SecularizationRate);
          end;
          if LgOrSmNow[K] = 'l' then
          begin
            MultFactor[K] := CloseNeighborCount[K]*(LgRelAttraction);
          end;
          if LgOrSmNow[K] = 'm' then
          begin
            MultFactor[K] := CloseNeighborCount[K]*(Trunc(((2*LgRelAttraction)+SmRelAttraction)/3));
          end;
          if LgOrSmNow[K] = 'k' then
          begin
            MultFactor[K] := CloseNeighborCount[K]*(Trunc((LgRelAttraction+(2*SmRelAttraction))/3));
          end;
          if LgOrSmNow[K] = 's' then
          begin
            MultFactor[K] := CloseNeighborCount[K]*(SmRelAttraction);
          end;
        end;
      end;
                        {3. REASSIGN, IF LARGEST MULTFACTOR IS HIGHER
                           THAN CURRENT INVESTMENT}

      if ChurchSizeNow = 'n' then InvMultiplier := NOInvRate;
      if ChurchSizeNow = 'l' then InvMultiplier := LgInvRate;
      if ChurchSizeNow = 'm' then InvMultiplier := Trunc(((2*LgInvRate)+SmInvRate)/3);
      if ChurchSizeNow = 'k' then InvMultiplier := Trunc((LgInvRate+(2*SmInvRate))/3);
      if ChurchSizeNow = 's' then InvMultiplier := SmInvRate;
      InvDivider := InvMultiplier;
      MostAttraction := 0;       {zeros size of largest group}
      MostAttractCode := 0;      {zeros code for identifying ties}
      RandomChoice := 1;
      for K := 0 to 11 do        {finds largest group of neighbors}
      begin
        if MultFactor[K] > MostAttraction then
        begin
          MostAttraction := MultFactor[K]
        end;
      end;
                       {changed for version 1.5}
      if IndivVary = false then MultipFactor := (InvMultiplier*InvMultiplier)*2/5 
                   else MultipFactor := (InvMultiplier*ConvTendArray[I,J])*2/5;
      if MostAttraction > MultipFactor then
      begin
        for K := 0 to 11 do    {checks for more than one largest group}
        begin
           if MultFactor[K] = MostAttraction
              then MostAttractCode := MostAttractCode + 1;
        end;
        if MostAttractCode > 1 then {chooses one of the ties at random}
        begin
          Randomize;
          RandomChoice := Random(MostAttractCode);
          if RandomChoice = 0 then RandomChoice := 1;
        end;
        L := 0;
        for K := 0 to 11 do
        begin
          if MultFactor[K] = MostAttraction then
          begin
            L := L +1;       {chooses which one of the ties to do}
            if L = RandomChoice then
            begin               {reassigns values}
                                {checks for barriers to conversion}
              if TreatAsBarrier = true then
              begin
                if LgOrSmNow[K] = 'n' then InvDivider := InvMultiplier;
                if LgOrSmNow[K] = 's' then InvDivider := SmInvRate;
                if LgOrSmNow[K] = 'k' then InvDivider := Trunc((LgInvRate+(2*SmInvRate))/3);
                if LgOrSmNow[K] = 'm' then InvDivider := Trunc(((2*LgInvRate)+SmInvRate)/3);
                if LgOrSmNow[K] = 'l' then InvDivider := LgInvRate;
              end;
                                {adjusts for barriers}
              if MultFactor[K] > (InvMultiplier*InvMultiplier*(InvDivider/InvMultiplier))*2/5 then
              begin
                M := 0;           {sorts through to pick right old value}
                while GridArray[I,J] <> ColorCode[M] do
                begin
                  M := M +1;
                end;
                                   {shifts church sizes, values}
                ChurchSize[M] := ChurchSize[M] - 1;
                GridArray[I,J] := ColorCode[K];
                ChurchSize[K] := ChurchSize[K] + 1;
                                   {shifts size code, if not 'w'}
                if M <> 0 then if ChurchSize[M] < ShiftToLarge then LgOrSmNow[M] := 'm';
                if M <> 0 then if ChurchSize[M] < ShiftToMed then LgOrSmNow[M] := 'k';
                if M <> 0 then if ChurchSize[M] < ShiftToSmall then LgOrSmNow[M] := 's';
                if K <> 0 then if ChurchSize[K] > ShiftToSmall then LgOrSmNow[K] := 'k';
                if K <> 0 then if ChurchSize[K] > ShiftToMed then LgOrSmNow[K] := 'm';
                if K <> 0 then if ChurchSize[K] > ShiftToLarge then LgOrSmNow[K] := 'l';
              end;
            end;
          end;
        end;
      end;



                        {4. FIGURE SCORE FOR PLAUSIBILITY}
                       {LESS SCORE FOR CURRENT INVESTMENT}
      PlausibilityCount := 0;
      for K := 0 to 11 do
      begin
        if CloseNeighborCount[K] <> 0
           then PlausibilityCount := PlausibilityCount + 1;
      end;
      if ((PlausibilityCount + PlausibilityRate - InvMultiplier) > 8) then
      begin
        if NOT (GridArray[I,J] = 'w') then
        begin
          L := 1;
          while GridArray[I,J] <> ColorCode[L] do
          begin
            L := L+1;
          end;
          ChurchSize[L] := ChurchSize[L] - 1;
          if ChurchSize[L] < ShiftToLarge then LgOrSmNow[L] := 'm';
          if ChurchSize[L] < ShiftToMed then LgOrSmNow[L] := 'k';
          if ChurchSize[L] < ShiftToSmall then LgOrSmNow[L] := 's';
          GridArray[I,J] := 'w';
          ChurchSize[0] := ChurchSize[0] + 1;
        end;
      end;
{  StopHere := true;}




    end;
  end;
{  StopHere := true;  }

                       {5. FIGURE ST. PAUL SCORE}
  if StPaulFactor = true then
  begin
    SeedRandom(StPaulOrder);
    for N := 1 to 100 do
    begin
         {find grid square corresponding to ORDER[N]}
      P := 1;
      if StPaulOrder[N] > 100*P then
      begin
        P := P + 1;
      end;
      P := P - 1;  {P = row}
      Q := StPaulOrder[N] - 100*P - 1;  {Q = col};
         {identify its current church}
      OldColor := GridArray[P,Q];
      NewColor := ColorCode[Random((ChurchCounter-1))];

         {shifts the old size counter}
      for K := 0 to ChurchCounter-1 do
      begin
        if GridArray[P,Q] = ColorCode[K] then
        begin
          ChurchSize[K] := ChurchSize[K]-1;
        end;
      end;

         {change Grid to one of the other religions & increases size counter}
      for K := 0 to ChurchCounter-1 do
      begin
        if ColorCode[K] = NewColor then
        begin
          GridArray[P,Q] := NewColor;
          ChurchSize[K] := ChurchSize[K] + 1;
        end;
      end;


    end;  {converts next one in order}
  end;

                        {REPAINT THE GRID}
 { StopHere := true;}

  GridForm.PaintScreen;
  ChurchForm.DisplaySize;
end;


{------------------------------ FORM PROCEDURES ----------------------------}

procedure TMainScreen.FormCreate(Sender: TObject);
var
  I, J : integer;
{  PanelHeight, PanelWidth: integer;}
begin
  ScreenHeight := Screen.Height;
  ScreenWidth := Screen.Width;
  if ScreenHeight < 800 then FillLbl1.Height := 150;
  if ScreenHeight < 700 then
  begin
                    {    LowResForm.Show;    }
    FillLbl1.Height := 100;
    FillLbl2.Height := 40;
    FillLbl3.Height := 30;
    FillLbl4.Height := 20;
    TitleLabel.Font.Size := 45;
    TitleLabel.Height := 75;
    DescriptionLbl.Font.Size := 20;
  end;

{                    THIS SECTION REJECTS 640x480 SCREEN RES.}
  if ScreenHeight < 550 then
  begin
     FillLbl1.Height := 50;
     TitleLabel.Caption := 'Screen Resolution Error!!';
     TitleLabel.Font.Size := 35;
     TitleLabel.Height := 50;
     DescriptionLbl.Font.Size := 20;
     DescriptionLbl.Caption := 'Simulation needs at least 800 by 600 pixel resolution.';
     FillLbl3.Font.Size := 20;
     FillLbl3.Caption := 'Close program, reset screen resolution, and try again.';
     AuthorLbl.Caption := ' ';
     CopyrightLbl.Caption := ' ';
     FreewareLbl.Caption := ' ';
     SetParameters.Enabled := false;
     Instructions1.Enabled := false;
  end;
                    {fix colors}
  SplashPanel.Color := clLtBlue;
  FreewareLbl.Color := clLtBlue;
  TitleLabel.Color := clLtBlue;
  DescriptionLbl.Color := clLtBlue;
  AuthorLbl.Color := clLtBlue;
  CopyrightLbl.Color := clLtBlue;
  EmailLbl.Color := clLtBlue;
  FillLbl1.Color := clLtBlue;
  FillLbl2.Color := clLtBlue;
  FillLbl3.Color := clLtBlue;
  FillLbl4.Color := clLtBlue;
  FillLbl5.Color := clLtBlue;

  TreatAsBarrier := false;
  IndivVary := false;
  PctSecular := 40;
  NoSmall := true;
  NoMed := true;
  NoLarge := true;
 for I := 0 to 99 do
  begin
    for J := 0 to 99 do BlankArray[I,J] := 'w';
  end;
  ColorCode[0] := 'w';
  ColorCode[1] := 'b';
  ColorCode[2] := 'g';
  ColorCode[3] := 'r';
  ColorCode[4] := 'y';
  ColorCode[5] := 'a';
  ColorCode[6] := 'l';
  ColorCode[7] := 'f';
  ColorCode[8] := 'm';
  ColorCode[9] := 'p';
  ColorCode[10] := 't';
  ColorCode[11] := 'n';


end;

procedure TMainScreen.Instructions2Click(Sender: TObject);
begin
  InstructionsForm.Show;
end;

procedure TMainScreen.SetParametersClick(Sender: TObject);
begin
  MainScreen.DescriptionLbl.Visible := true;
  MainScreen.TitleLabel.Visible := true;
  MainScreen.AuthorLbl.Visible := true;
  MainScreen.EmailLbl.Visible := true;
  MainScreen.FreewareLbl.Visible := true;
  MainScreen.CopyrightLbl.Visible := true;

{  ChurchForm.FormStyle := fsNormal;}
  ChurchForm.Visible := false;
{  GridForm.FormStyle := fsNormal;}
  GridForm.Visible := false;
  SettingsForm.Visible := false;
  SeedRandom(Order);
  with Parameters do
  begin
    if (NumLgRelsBox.Text = '0') AND
       (NumSmRelsBox.Text = '0')
    then
    begin
      OKBtn.Enabled := false;
    end;
  end;
  Parameters.Show;

end;

procedure TMainScreen.RunOneClick(Sender: TObject);
begin
  Reassign;
end;


procedure TMainScreen.RunFiveClick(Sender: TObject);
var
  K : integer;
begin
  K := 0;
  Reassign;
  K := K + 1;
  Reassign;
  K := K + 1;
  Reassign;
  K := K + 1;
  Reassign;
  K := K + 1;
  Reassign;
  K := K + 1;
end;

procedure TMainScreen.RunTenClick(Sender: TObject);
var
  K : integer;
begin
  K := 0;
  Reassign;
  K := K + 1;
  Reassign;
  K := K + 1;
  Reassign;
  K := K + 1;
  Reassign;
  K := K + 1;
  Reassign;
  K := K + 1;
  Reassign;
  K := K + 1;
  Reassign;
  K := K + 1;
  Reassign;
  K := K + 1;
  Reassign;
  K := K + 1;
  Reassign;
  K := K + 1;
end;

procedure TMainScreen.Run50Click(Sender: TObject);
begin
  RunTenClick(MainScreen);
  RunTenClick(MainScreen);
  RunTenClick(MainScreen);
  RunTenClick(MainScreen);
  RunTenClick(MainScreen);
end;


procedure TMainScreen.ResetSimulationClick(Sender: TObject);
begin

  GridForm.ClearGrid;
  GridArray := StartArray;
  ChurchSize := StartChurchSize;
  LgOrSmNow := StartLgOrSmNow;
  ChurchForm.DisplaySize;
  GridForm.PaintScreen;
end;

procedure TMainScreen.ExitClick(Sender: TObject);
begin
{  ChurchForm.Close;
  GridForm.Close;
  Close; }
{  Application.Terminate;}
  ExitBox.Show;

end;

procedure TMainScreen.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Application.Terminate;
end;





procedure TMainScreen.ProgramOptions1Click(Sender: TObject);
begin
  OptionsForm.Show;
end;

procedure TMainScreen.Background1Click(Sender: TObject);
begin
  BackGroundForm.Show;
end;

end.
